/*
 *	fonts.c
 */

#include "filer.h"

typedef struct _FONTS {
	u8		fontname[XF_NAMELEN+1];
	int		fontsize;
	PXFONTHEAD	pxa;
	PXFONTHEAD	pxk;
} FONTS, *PFONTS;

static PFONTS	pFonts;
static int	nFonts;

static int	CurFontIndex;

static u8	CurFontPitch;

/*
 *	IsValidFontInfo
 */
static BOOL IsValidFontInfo(PFONTINFO pfi)
{
	if (memcmp((u8 *)pfi->sig, FI_SIG, sizeof(pfi->sig)) ||
		pfi->ver != FI_VER ||
		pfi->size != sizeof(*pfi))
		return FALSE;
	return TRUE;
}

/*
 *	LoadFontName
 */
static void LoadFontName(u8 *pfontname, int *pfontsize)
{
	PFONTINFO	pfi;

	pfi = (PFONTINFO)FONTINFO_ADDR;
	if (!IsValidFontInfo(pfi)) {
		memset(pfontname, 0, XF_NAMELEN + 1);
		*pfontsize = 0;
		return;
	}

	ReadSRAMByte(pfi->fontname, pfontname, XF_NAMELEN + 1);
	*pfontsize = ReadSRAM8(&pfi->fontsize);
}

/*
 *	LoadFontPitch
 */
static void LoadFontPitch(u8 *pfontpitch)
{
	PFONTINFO	pfi;

	pfi = (PFONTINFO)FONTINFO_ADDR;
	if (!IsValidFontInfo(pfi)) {
		*pfontpitch = FP_DEFAULT;
		return;
	}

	*pfontpitch = ReadSRAM8(&pfi->fontpitch);
}

/*
 *	BuildFontInfo
 */
static void BuildFontInfo(PFONTINFO pfi)
{
	static const	zeroname[XF_NAMELEN + 1];

	WriteSRAMByte(pfi->sig, FI_SIG, sizeof(pfi->sig));
	WriteSRAM8(&pfi->ver, FI_VER);
	WriteSRAM8(&pfi->size, sizeof(*pfi));

	WriteSRAMByte(pfi->fontname, (u8 *)zeroname, XF_NAMELEN + 1);
	WriteSRAM8(&pfi->fontsize, 0);
	WriteSRAM8(&pfi->fontpitch, FP_DEFAULT);
}

/*
 *	SaveFontName
 */
static void SaveFontName(u8 *pfontname, int fontsize)
{
	PFONTINFO	pfi;

	pfi = (PFONTINFO)FONTINFO_ADDR;
	if (!IsValidFontInfo(pfi))
		BuildFontInfo(pfi);

	WriteSRAMByte(pfi->fontname, pfontname, XF_NAMELEN + 1);
	WriteSRAM8(&pfi->fontsize, fontsize);
}

/*
 *	SaveFontPitch
 */
static void SaveFontPitch(u8 fontpitch)
{
	PFONTINFO	pfi;

	pfi = (PFONTINFO)FONTINFO_ADDR;
	if (!IsValidFontInfo(pfi))
		BuildFontInfo(pfi);

	WriteSRAM8(&pfi->fontpitch, fontpitch);
}

/*
 *	FindFont
 */
static int FindFont(u8 *pfontname, int fontsize)
{
	int		i;
	PFONTS		pf;

	for (i = 0, pf = pFonts; i < nFonts; i++, pf++) {
		if (strcmp(pf->fontname, pfontname) == 0 &&
			pf->fontsize == fontsize)
			break;
	}

	return i;
}

/*
 *	InitFonts
 */
void InitFonts()
{
	PDIRENT		pdir;
	PDIRENT		pd, pe;
	PXFONTHEAD	pxf;
	PFONTS		pf, pf2;
	FONTS		ftmp;
	int		i, j;
	u8		*ptop, *pend;
	u8		fontname[XF_NAMELEN + 1];
	int		fontsize;

	if (!(pdir = FsGetDirEnt((u8 *)FONT_PATH))) {
		Fix8PutStr(10, 80, "InitFonts: Can't find font directory.");
		for (;;) ;
	}

	XfontInit();

	/*
	 *	Get all fonts
	 */

	iGetHeap(&ptop, &pend);
	pFonts = (PFONTS)ptop;
	nFonts = 0;

	pd = (PDIRENT)(pdir->fileaddr + pdir->namesize);
	pe = (PDIRENT)((u8 *)pd + pdir->filesize);
	for (; pd < pe; pd++) {
		if (pd->attr & DA_DIR)
			continue;
		pxf = (PXFONTHEAD)(pd->fileaddr + pd->namesize);
		if (!IsValidXfont(pxf))
			continue;

		for (i = 0, pf = pFonts; i < nFonts; i++, pf++) {
			if (strcmp(pf->fontname, pxf->name) == 0 &&
				pf->fontsize == pxf->height)
				break;
		}
		if (&pFonts[i] >= (PFONTS)pend)
			break;
		if (i >= nFonts) {
			memcpy(pf->fontname, pxf->name, sizeof(pf->fontname));
			pf->fontsize = pxf->height;
			nFonts++;
		}

		if (ISDBCSFONT(pxf))
			pf->pxk = pxf;
		else
			pf->pxa = pxf;
	}

	iSetHeapTop((u8 *)(&pFonts[nFonts]));

	/*
	 *	Check fonts
	 */

	if (nFonts <= 0) {
		Fix8PutStr(10, 80, "InitFonts: fonts not found.");
		for (;;) ;
	}

	for (i = 0, pf = pFonts; i < nFonts - 1; i++, pf++)
		if (!pf->pxa) {
			Fix8PutStr(10, 80, "InitFonts: No alphabet font");
			for (;;) ;
		}

	/*
	 *	Sort fonts
	 */

	for (i = 0, pf = pFonts; i < nFonts - 1; i++, pf++)
		for (j = i + 1, pf2 = pf + 1; j < nFonts; j++, pf2++)
			if (pf->fontsize > pf2->fontsize ||
				(pf->fontsize == pf2->fontsize &&
				pf->pxk->width + pf->pxk->xgap >
				pf2->pxk->width + pf2->pxk->xgap))
			{
				ftmp = *pf;
				*pf = *pf2;
				*pf2 = ftmp;
			}

	/*
	 *	Load font informations
	 */

	LoadFontPitch(&CurFontPitch);
	XfontSetFontPitch(CurFontPitch);

	pf = &pFonts[0];	// Smallest font
	XfontSetFontObject(FO_SMALL, pf->pxa, pf->pxk);

	LoadFontName(fontname, &fontsize);
	CurFontIndex = FindFont(fontname, fontsize);
	if (CurFontIndex >= nFonts)
		CurFontIndex = 0;

	pf = &pFonts[CurFontIndex];
	XfontSetFontObject(FO_NORMAL, pf->pxa, pf->pxk);

	XfontSelectFontObject(FO_NORMAL);
}

#define PREVIEW_GAP	8
#define PREVIEW_H	32

/*
 *	PreviewFont
 */
static int PreviewFont(int index, void *params)
{
	PFONTS		pf;
	PRECT		prc;
	RECT		rc;
	int		w, h;

	if (index < 0 || index >= nFonts)
		return -1;
	prc = (PRECT)params;

	pf = &pFonts[index];
	XfontSetFontObject(FO_USER, pf->pxa, pf->pxk);
	XfontSelectFontObject(FO_USER);

	w = XfontGetStrWidth((u8 *)sPreviewSample);
	h = XfontGetCharHeight();

	EraseScrn(prc->x, prc->y, prc->w, prc->h);
	LeftPrintf(prc->x, prc->y, prc->w, "%s", sPreviewSample);
	if (pf->pxk)
		LeftPrintf(prc->x + w, prc->y, prc->w - w, "%s",
			sPreviewKanjiSample);

	return 0;
}

/*
 *	SelectFont
 */
BOOL SelectFont(u8 *title)
{
	int		i;
	PFONTS		pf;
	LISTBOX		lb;
	u8		**ppitems;
	u8		*pitems;
	u8		*ptop, *pend;
	int		n;
	int		res = FALSE;
	RECT		rc;
	RECT		rcpv;

	/*
	 *	Choose font
	 */

	iGetHeap(&ptop, &pend);
	ppitems = (u8 **)ptop;
	pitems = &ptop[sizeof(ppitems[0]) * nFonts];
	if (pitems >= pend)
		return FALSE;

	memset(&lb, 0, sizeof lb);
	lb.title = title;
	lb.nitems = nFonts;
	lb.defitem = CurFontIndex;
	for (i = 0; i < nFonts; i++) {
		ppitems[i] = pitems;
		if (&pitems[XF_NAMELEN + 3] >= pend)
			return FALSE;
		n = _sprintf(ppitems[i], "%s%3d",
			pFonts[i].fontname, pFonts[i].fontsize);
		pitems += (n + 1);	// including nul
	}
	lb.items = ppitems;
	iSetHeapTop(pitems);

	GetClientRect(&rc);
	lb.rc.x = rc.x;
	lb.rc.y = rc.y;
	lb.rc.w = rc.w;
	lb.rc.h = rc.h - (PREVIEW_GAP + PREVIEW_H);

	rcpv.x = rc.x;
	rcpv.y = rc.y + lb.rc.h + PREVIEW_GAP;
	rcpv.w = rc.w;
	rcpv.h = PREVIEW_H;

	lb.callback = PreviewFont;
	lb.params = (void *)&rcpv;

	i = ListBox(&lb);

	if (i < 0 || i == CurFontIndex)
		goto out;

	/*
	 *	Select font
	 */

	CurFontIndex = i;
	pf = &pFonts[CurFontIndex];

	XfontSetFontObject(FO_NORMAL, pf->pxa, pf->pxk);

	SaveFontName(pf->fontname, pf->fontsize);

	res = TRUE;

out:
	iSetHeapTop(ptop);

	return res;
}

static const u8		iFontPitch[] = {
	FP_DEFAULT,
	0,
	1,
	2,
	3,
	4,
	5,
};

#define nFontPitch	((int)arraysizeof(iFontPitch))

/*
 *	FindPitchIndex
 */
static int FindPitchIndex(u8 pitch)
{
	int		i;

	for (i = 0; i < nFontPitch; i++)
		if (iFontPitch[i] == pitch)
			return i;

	return 0;
}

/*
 *	PreviewFontPitch
 */
static int PreviewFontPitch(int index, void *params)
{
	PRECT		prc;
	PXFONTOBJ	pxf;
	int		w, h;
	int		opitch;

	if (index < 0 || index >= nFontPitch)
		return -1;
	prc = (PRECT)params;

	XfontSelectFontObject(FO_SMALL);
	opitch = XfontSetFontPitch(iFontPitch[index]);
	pxf = XfontGetFontObject(0x81);

	w = XfontGetStrWidth((u8 *)sPreviewSample);
	h = XfontGetCharHeight();

	EraseScrn(prc->x, prc->y, prc->w, prc->h);

	LeftPrintf(prc->x, prc->y, prc->w, "%s", sPreviewSample);
	if (pxf)
		LeftPrintf(prc->x + w, prc->y, prc->w - w, "%s",
			sPreviewKanjiSample);

	LeftPrintf(prc->x, prc->y + h, prc->w, "%s", sPreviewSample);
	if (pxf)
		LeftPrintf(prc->x + w, prc->y + h, prc->w - w, "%s",
			sPreviewKanjiSample);

	XfontSetFontPitch(opitch);

	return 0;
}

/*
 *	SelectFontPitch
 */
int SelectFontPitch(u8 *title)
{
	LISTBOX		lb;
	int		i;
	int		index;
	int		h;
	u8		*apitch[nFontPitch];
	RECT		rc;
	RECT		rcpv;

	memset(&lb, 0, sizeof lb);
	lb.title = title;
	lb.nitems = nFontPitch;
	index = FindPitchIndex(CurFontPitch);
	lb.defitem = index;
	apitch[0] = "Default";
	apitch[1] = "0";
	apitch[2] = "1";
	apitch[3] = "2";
	apitch[4] = "3";
	apitch[5] = "4";
	apitch[6] = "5";
	lb.items = apitch;

	XfontSelectFontObject(FO_SMALL);
	h = XfontGetCharHeight();
	GetClientRect(&rc);
	lb.rc.x = rc.x;
	lb.rc.y = rc.y;
	lb.rc.w = rc.w;
	lb.rc.h = (h + 2) * nFontPitch;

	rcpv.x = rc.x;
	rcpv.y = rc.y + lb.rc.h;
	rcpv.w = rc.w;
	rcpv.h = rc.h - lb.rc.h;

	lb.callback = PreviewFontPitch;
	lb.params = (void *)&rcpv;

	i = ListBox(&lb);

	if (i < 0)
		return i;
	if (i == index)
		return LB_CANCEL;

	CurFontPitch = iFontPitch[i];
	XfontSetFontPitch(CurFontPitch);

	SaveFontPitch(CurFontPitch);

	return i;
}

